@@ -211,6 +211,14 @@ class ShellTest(testutils.TestCase):
             self.assertEqual(0, actual)
             self.assertFalse(et_mock.called)
 
+    def test_help_no_subcommand(self):
+        shell = openstack_shell.OpenStackImagesShell()
+        argstr = '--os-image-api-version 2'
+        with mock.patch.object(shell, '_get_keystone_auth_plugin') as et_mock:
+            actual = shell.main(argstr.split())
+            self.assertEqual(0, actual)
+            self.assertFalse(et_mock.called)
+
     def test_blank_call(self):
         shell = openstack_shell.OpenStackImagesShell()
         with mock.patch.object(shell, '_get_keystone_auth_plugin') as et_mock:
@@ -592,6 +600,17 @@ class ShellTest(testutils.TestCase):
         self.assertEqual(glance_logger.getEffectiveLevel(), logging.DEBUG)
         conf.assert_called_with(level=logging.DEBUG)
 
+    def test_subcommand_help(self):
+        # Ensure that main works with sub command help
+        stdout, stderr = self.shell('help stores-delete')
+
+        expected = 'usage: glance stores-delete --store <STORE_ID> ' \
+                   '<IMAGE_ID>\n\nDelete image from specific store.' \
+                   '\n\nPositional arguments:\n  <IMAGE_ID>          ' \
+                   'ID of image to update.\n\nRequired arguments:\n  ' \
+                   '--store <STORE_ID>  Store to delete image from.\n'
+        self.assertEqual(expected, stdout)
+
 
 class ShellTestWithKeystoneV3Auth(ShellTest):
     # auth environment to use
@@ -692,7 +711,7 @@ class ShellTestWithNoOSImageURLPublic(ShellTestWithKeystoneV3Auth):
         glance_shell = openstack_shell.OpenStackImagesShell()
         glance_shell.main(args.split())
         self.assertEqual(self.requests.request_history[2].url,
-                         self.image_url + "v2/images?limit=20&"
+                         self.image_url + "v2/images?limit=200&"
                          "sort_key=name&sort_dir=asc")
 
 
